# Copyright 2018 Heisenberg Quantum Simulations
# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""This package provdes a port between Cirq and ProejctQ.

Cirq is an open source quantum computing framework developed by Google and
Projectq is a framework to compile and simulate quantum algorithms.

Cirqprojectq provides two main functionalities: Firstly, it provides methods and
classes that can port algorithms from projectq to cirq and secondly it introduces
Xmon gates and xmon gate compositions to projectq.

.. autosummary::

    xmon_gates
    xmon_decompositions
    xmon_rules
    xmon_setup
    circ_engine
"""

from . import (xmon_gates,
               xmon_decompositions,
               xmon_rules,
               xmon_setup,
               circ_engine)